#ifndef CLASSIFIER_SUPPORT_H
#define CLASSIFIER_SUPPORT_H

/*******************************************************************************
@ddblock_begin copyright

Copyright (c) 2017-2020
UMB-UMD Neuromodulation Research Group,
University of Maryland at Baltimore, and 
University of Maryland at College Park. 

All rights reserved.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BALTIMORE
OR UNIVERSITY OF MARYLAND COLLEGE PARK BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@ddblock_end copyright
*******************************************************************************/

//Standard headers
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

//Data definitions
#define FEATURE_LENGTH	200
#define NUM_CLASSES		2

//Classifier definitions
#define ONE_VS_ONE	1
#ifdef ONE_VS_ONE	//For one vs one, number of learners is k*(k-1)/2 for k classes
	#define NUM_LEARNERS	((NUM_CLASSES)*(NUM_CLASSES-1))/2
#else				//For one vs all, number of learners is k, for k classes
	#define NUM_LEARNERS	NUM_CLASSES
#endif
	
/**
 * Data structure containing the classifier parameters
 * Weights  : Learner Weights
 * Means    : Learner means
 * Std_Devs : Standard deviations of each learner
 * Bias		: Bias for each learner
 * Codmat	: Coding matrix for combining learners
 */
typedef struct params
{
		double *Weights;	//Learner Weights
		double *Means;		//Learner means
		double *Std_Devs;	//Learner sigmas
		double *Bias;		//Bias 
		double *Codmat;		//Coding matrix combining the learners
}Classifier_params_t,*Classifier_params_ptr;


/** 
 *	@brief Computes the hinge loss value of the given class score
 *
 *  @param score Score to compute hinge loss for
 *  @return hinge The hinge loss computed
 */
double hinge(double score);


/** 
 *	@brief Returns the index of the maximum value in array
 *
 *  @param array Array to be iterated over
 *	@param length No. of samples in array
 *  @return max index of maximum value
 */
uint32_t max(double* array, uint32_t length);


/**
 *	@brief Performs SVM classification on feature vector 'X' using SVM parameters
 *
 *  @param  X  feature vector
 *  @param  Params  Pointer to the classifier parameters structure
 *	@return prediction: Predicted class
 */
uint8_t classify(double* X, Classifier_params_ptr Params);

/** 
 *	@brief Subtracts elements of b from corresponding elements of a
 *
 *  @param a pointer to first vector
 *  @param b pointer to second vector
 *  @param length number of samples in each vector
 *  @return void
 */
void vector_subtract_inplace(double *a, double *b, uint32_t length);


/** 
 *	@brief Divides elements of a by corresponding elements of b
 *
 *  @param a pointer to first vector
 *  @param b pointer to second vector
 *  @param length number of samples in each vector
 *  @return void
 */
void vector_scale_inplace(double *a, double *b, uint32_t length);


/** 
 *	@brief Computes the dot product c of vectors a and b 
 *
 *  @param a pointer to first vector
 *  @param b pointer to second vector
 *  @param c pointer to destination vector 
 *  @param length number of samples in each vector
 *  @return void
 */
void vector_dotproduct(double *a, double *b, double *c, uint32_t length);	

#endif
