#ifndef _welt_cpp_actor_h
#define _welt_cpp_actor_h

/*******************************************************************************
@ddblock_begin copyright

Copyright (c) 1997-2020
Maryland DSPCAD Research Group, The University of Maryland at College Park 
All rights reserved.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@ddblock_end copyright
*******************************************************************************/
#include "welt_cpp_graph.h"

/* This is an abstract base class for the run-time context of an actor instance.
 * */
class welt_cpp_actor{
/* Common elements across context type of actors. */
protected:
    /* The mode of an actor */
    int mode;
    /* The index of an actor */
    int index;
    /* The number of ports of an actor */
    int port_count;
    /* The maximum number of ports of an actor */
    int max_port_count;

public:
    /* The members void **paramrefs, int param_count, int *param_length, int
     * max_param_count are replaced by the following params_vec vector. The
     * params_vec vector stores the pointers of parameters that should be
     * initialized and allocated in an actor. */
    vector<void*> params_vec;

    /* The enable method of a welt-cpp actor corresponds to the enable function
     * that enables an actor of the lightweight/welterweight dataflow API. Every
     * actor must have an enable function. */
    virtual boolean enable() = 0;

    /* The invoke method of a welt-cpp actor corresponds to the invoke function
     * that invokes an actor of the lightweight/welterweight dataflow API. Every
     * actor must have an invoke function. */
    virtual void invoke() = 0;

    /* The reset method of a welt-cpp actor corresponds to the reset function
     * that resets an actor of the lightweight/welterweight dataflow API. Every
     * actor must have an reset function. */
    virtual void reset() = 0;

    /* The connect method of a welt-cpp actor corresponds to the connect
     * function that adds an actor to a dataflow graph of the
     * lightweight/welterweight API. Every actor must have an connect function.
     * */
//    has to include welt_cpp_graph.h in this .h file
    virtual void connect(welt_cpp_graph *graph) = 0;

    /* The destructor of a welt-cpp actor corresponds to the terminate function
     * of the lightweight/welterweight dataflow API. Every actor must have a
     * terminate function. This base class destructor does nothing, and just
     * serves to ensure that all actors have terminate functions even if they
     * are not explicitly defined for the concrete classes. */
    virtual ~welt_cpp_actor() = default;

    /* A collection of subgraph pointers */
    vector<welt_cpp_graph*> subgraphs;

    /* The number of subgraphs */
    int subgraph_count;

    /* FIXME: port the following member */

    vector<lide_c_fifo_pointer*> portrefs;

    int welt_cpp_actor_get_index();
    void welt_cpp_actor_set_index(int index);
    void welt_cpp_actor_set_portref(int index, lide_c_fifo_pointer fifo);
    lide_c_fifo_pointer welt_cpp_actor_get_portref(int index);
    void welt_cpp_actor_init_vari_portrefs(int count, ...);
    void welt_cpp_actor_init_portrefs();
    void welt_cpp_actor_add_portrefs(lide_c_fifo_pointer *port);
    void welt_cpp_actor_add_portrefs_array(lide_c_fifo_pointer *port_array, int length);
    void welt_cpp_actor_reset_portrefs();
    void welt_cpp_actor_set_paramref(int param_index, void *data);
    void welt_cpp_actor_get_paramref(int param_index, void *data);
    void welt_cpp_actor_set_param_length(int param_index, int length);
    int welt_cpp_actor_get_param_length(int param_index);
    void welt_cpp_actor_init_vari_paramrefs(int *data_length,int count, ...);
    void welt_cpp_actor_init_paramrefs();
    void welt_cpp_actor_add_paramtrefs_array(void *data_array,int *data_length, int length);
    void welt_cpp_actor_add_paramrefs(void *data, int data_length);
    void welt_cpp_actor_reset_paramrefs();
    void welt_cpp_actor_setInvoke(void invoke());
    void welt_cpp_actor_invoke_diagnostic();
    void lide_c_connect_fifo(int port_index, lide_c_fifo_pointer* fifo_pointer);
};

#endif //_welt_cpp_actor_h
