/*******************************************************************************
@ddblock_begin copyright

Copyright (c) 1997-2020
Maryland DSPCAD Research Group, The University of Maryland at College Park 
All rights reserved.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@ddblock_end copyright
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "lide_c_inner_product.h"
#include "lide_c_util.h"
#include "lide_c_graph.h"

#define MAX_FIFO_COUNT  4

/*******************************************************************************
INNER PRODUCT STRUCTURE DEFINITION
*******************************************************************************/

struct _lide_c_inner_product_context_struct { 
#include "lide_c_actor_context_type_common.h"

    /* Vector length variable. */
    int length;  

    /* Input ports. */
    lide_c_fifo_basic_pointer m;
    lide_c_fifo_basic_pointer x;
    lide_c_fifo_basic_pointer y;

    /* Output port. */
    lide_c_fifo_basic_pointer out;
};

/*******************************************************************************
IMPLEMENTATIONS OF INTERFACE FUNCTIONS.
*******************************************************************************/

lide_c_inner_product_context_type *lide_c_inner_product_new(
        lide_c_fifo_pointer m, lide_c_fifo_pointer x, lide_c_fifo_pointer y,
        lide_c_fifo_pointer out, int index) { 

    lide_c_inner_product_context_type *context = NULL;

    context = lide_c_util_malloc(sizeof(lide_c_inner_product_context_type));
    context->mode = LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH;
    context->enable =
            (lide_c_actor_enable_ftype)lide_c_inner_product_enable;
    context->invoke = 
            (lide_c_actor_invoke_ftype)lide_c_inner_product_invoke;
    context->reset = (lide_c_actor_reset_ftype)lide_c_inner_product_reset;
    context->length = 0;
    context->m = (lide_c_fifo_basic_pointer)m;
    context->x = (lide_c_fifo_basic_pointer)x;
    context->y = (lide_c_fifo_basic_pointer)y;
    context->out = (lide_c_fifo_basic_pointer)out;
    
    context->index = index;
    context->connect = (lide_c_actor_connect_ftype)lide_c_inner_product_connect;
    context->max_port_count = MAX_FIFO_COUNT;
    
    
    lide_c_actor_init_portrefs((lide_c_actor_context_type *)context);
    lide_c_actor_add_portrefs((lide_c_actor_context_type *)context, 
                                (lide_c_fifo_pointer *)&context->m);    
    lide_c_actor_add_portrefs((lide_c_actor_context_type *)context, 
                                (lide_c_fifo_pointer *)&context->x); 
    lide_c_actor_add_portrefs((lide_c_actor_context_type *)context, 
                                (lide_c_fifo_pointer *)&context->y);  
    lide_c_actor_add_portrefs((lide_c_actor_context_type *)context, 
                                (lide_c_fifo_pointer *)&context->out); 
    
    
    return context;
}

boolean lide_c_inner_product_enable(
        lide_c_inner_product_context_type *context) {
    boolean result = FALSE;

    switch (context->mode) {
    case LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH:
        result = lide_c_fifo_basic_population(context->m) >= 1;
        break;
    case LIDE_C_INNER_PRODUCT_MODE_PROCESS:
        result = (lide_c_fifo_basic_population(context->x) >= context->length) 
                    && (lide_c_fifo_basic_population(context->y) 
                    >= context->length) && 
                    (lide_c_fifo_basic_population(context->out) <
                    lide_c_fifo_basic_capacity(context->out));
        break;
    default:
            result = FALSE;
            break;
    }
    return result;
}

void lide_c_inner_product_invoke(lide_c_inner_product_context_type *context) {
    int i = 0;
    int sum = 0;
    int x_value = 0;
    int y_value = 0;

    switch (context->mode) {
    case LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH:
        lide_c_fifo_basic_read(context->m, &(context->length));
        if (context->length <= 0) {
            context->mode = LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH;
            return;
        }
        context->mode = LIDE_C_INNER_PRODUCT_MODE_PROCESS;
        break;

    case LIDE_C_INNER_PRODUCT_MODE_PROCESS:
        for (i = 0; i < context->length; i++){
            lide_c_fifo_basic_read(context->x, &(x_value));
            lide_c_fifo_basic_read(context->y, &(y_value));
            sum += (x_value * y_value);
        }
        lide_c_fifo_basic_write(context->out, &sum);
        context->mode = LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH;
        break; 
    default:
        context->mode = LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH;
        break;
    } 
    return;
}

void lide_c_inner_product_terminate(
        lide_c_inner_product_context_type *context) {
    free(context);
}

void lide_c_inner_product_reset(lide_c_inner_product_context_type *context){
    context->mode = LIDE_C_INNER_PRODUCT_MODE_STORE_LENGTH;
    context->length = 0;
    return;
}


void lide_c_inner_product_connect(lide_c_actor_context_type *context, 
        struct lide_c_graph_context_struct *graph){
    int port_index;
    int direction;
    /* input 1*/
    direction = GRAPH_IN_CONN_DIRECTION;
    port_index = LIDE_C_IP_PORT_M;
    lide_c_graph_add_connection(graph, (lide_c_actor_context_type *)context, 
                                port_index, direction);
    
    /* input 2*/
    direction = GRAPH_IN_CONN_DIRECTION;
    port_index = LIDE_C_IP_PORT_X;
    lide_c_graph_add_connection(graph, (lide_c_actor_context_type *)context, 
                                port_index, direction);
    
    /* input 3*/
    direction = GRAPH_IN_CONN_DIRECTION;
    port_index = LIDE_C_IP_PORT_Y;
    lide_c_graph_add_connection(graph, (lide_c_actor_context_type *)context, 
                                port_index, direction);
    
    /* output 1*/
    direction = GRAPH_OUT_CONN_DIRECTION;
    port_index = LIDE_C_IP_PORT_OUT;
    lide_c_graph_add_connection(graph, (lide_c_actor_context_type *)context, 
                                port_index, direction);
    return;
} 
