/****************************************************************************
@ddblock_begin copyright

Copyright (c) 1997-2017
Maryland DSPCAD Research Group, The University of Maryland at College Park 

Permission is hereby granted, without written agreement and without
license or royalty fees, to use, copy, modify, and distribute this
software and its documentation for any purpose, provided that the above
copyright notice and the following two paragraphs appear in all copies
of this software.

IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
ENHANCEMENTS, OR MODIFICATIONS.

@ddblock_end copyright
******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lide_c_util.h"
#include "lide_c_router_v2.h"


struct _lide_c_router_v2_context_struct {
    /* The number of tokens that all the buffers can hold. */
    int capacity;

    /*  Pointer to an array of buffer start pointers; each buffer start 
        pointer points to the first token placeholder in the buffer*/
    void **buffer_start;

    /*  Pointer to an array of buffer end pointers;
        Pointer to the last token placeholder in the buffer. This member
        is here for convenience (it can also be derived from the 
        buffer start, capacity, and token_size). 
    */
    void **buffer_end;

    /*  The number of tokens that are currently in all the buffers of the 
        router. */
    int population;

    /*  Pointer to an array of read pointers;
        Pointer to the next token placeholider to be read from. */
    void **read_pointer;

    /*  Pointer to an array of write pointers;
        Pointer to the next token placeholider to be written into. */
    void **write_pointer;

    /*  The number of bytes in a single token. */
    int token_size;

    /* router port array*/
    lide_c_router_v2_port_pointer *ports;

    /* input_ports_count*/
    int input_ports_count;

    /* output_ports_count*/
    int output_ports_count;

    /* port counts*/
    int ports_count;

    /* index of the router in a dataflow graph. It is flexible and set by the 
    designer*/
    int index;
    
    /* enabled buffer id for write operations*/
    int en_buff_id;
    
    /* methods in router*/
    lide_c_router_v2_read_ftype read;
    lide_c_router_v2_peek_ftype peek;
    lide_c_router_v2_read_advance_ftype read_advance;
    lide_c_router_v2_read_block_ftype read_block;
    lide_c_router_v2_write_ftype write;
    //lide_c_router_write_advance_ftype write_advance;
    lide_c_router_v2_write_block_ftype write_block;

    lide_c_router_v2_context_population_ftype get_cpop;
    lide_c_router_v2_port_population_ftype get_ppop;
    lide_c_router_v2_population_ftype get_pop;

    lide_c_router_v2_capacity_ftype get_cap;
    lide_c_router_v2_context_capacity_ftype get_ccap;
    lide_c_router_v2_port_capacity_ftype get_pcap;

    lide_c_router_v2_token_size_ftype  get_token_size;
    lide_c_router_v2_reset_ftype reset;
    lide_c_router_v2_free_ftype free;


    lide_c_router_v2_get_router_port_with_index_ftype get_port_withidx;
    lide_c_router_v2_get_router_port_ftype get_port;


    /* The token printing function (if any) associated with this router. */
    lide_c_router_v2_f_token_printing print; 
};

struct _lide_c_router_v2_port_struct {
    
    /*  Each buffer start pointer points to the first token placeholder 
        in the buffer*/
    void *buffer_start;

    /* Pointer to the last token placeholder in the buffer. This member
       is here for convenience (it can also be derived from the 
       buffer start, capacity, and token_size). 
    */
    void *buffer_end;

    /*  Pointer to the next token placeholider to be read from. */
    void **read_pointer;

    /*  Pointer to the next token placeholider to be written into. */
    void **write_pointer;

    /* label for input or output port group*/
    int group_label;    
    /* index in the group*/
    int group_idx;    
    /* index in the router */    
    int index;   
    /* Port's current "population" for every port*/    
    int population;  
    /* Router's current "capacity" for every port*/    
    int capacity;   
    /* Read pointer for an output port; set to NULL if input port */
    //void read_pointer;  
    /* Router context */
    lide_c_router_v2_context_pointer  router_context; 
    /* Get context*/
    lide_c_router_v2_port_get_router_context_ftype get_context;
    lide_c_router_v2_port_population_ftype get_port_pop;
    lide_c_router_v2_port_get_index_ftype get_port_index;
    
};


/*****************************************************************************
Public functions.
*****************************************************************************/

/* router port construct function */
/*  input port label:INPUT_PORT --> 0; 
    output port label: OUTPUT_PORT --> 1 */
lide_c_router_v2_port_pointer lide_c_router_v2_port_new(
                                        lide_c_router_v2_context_pointer router, 
                                        int capacity,
                                        int group_label, int group_idx, 
                                        int index){
    lide_c_router_v2_port_type *port = NULL;

    port = lide_c_util_malloc(sizeof(lide_c_router_v2_port_type));
    
    port->buffer_start = router->buffer_start[group_idx];
    port->buffer_end = router->buffer_end[group_idx];
    port->read_pointer = &router->read_pointer[group_idx];
    port->write_pointer = &router->write_pointer[group_idx];
    
    port->group_label = group_label;
    port->group_idx = group_idx;
    port->population = 0;
    port->capacity = capacity;
    /*
    port->index = index + router->input_ports_count * port->group_label + 
                    port->group_idx;
    */
    port->index = index;
    port->router_context = router;
    port->get_context = lide_c_router_v2_port_get_router_context;
    port->get_port_pop = lide_c_router_v2_port_population;
    port->get_port_index = lide_c_router_v2_port_get_index;
    return port;
}

lide_c_router_v2_context_pointer lide_c_router_v2_context_new(
                                                        int port_capacity, 
                                                        int token_size,
                                                        int input_ports_count,
                                                        int output_ports_count,
                                                        int index){
    int i;
    int tmp;
    
    lide_c_router_v2_context_pointer context = NULL;
    
    if (sizeof(char) != 1) {
        fprintf(stderr, "incompatible target platform\n");
    }

    if (port_capacity < 1) {
        return NULL;
    }
    if (token_size <= 0) {
        return NULL;
    }

    context = (lide_c_router_v2_context_pointer)lide_c_util_malloc(
                sizeof(lide_c_router_v2_context_type));
    context->population = 0;
    
    context->token_size = token_size;
    context->input_ports_count = input_ports_count;
    context->output_ports_count = output_ports_count;
    context->capacity = port_capacity * context->output_ports_count;
    context->ports_count = context->input_ports_count + 
                            context->output_ports_count;
    context->index = index;
    context->en_buff_id = 0;
    /* Memory initialization*/
    context->buffer_start = NULL;
    context->buffer_start = (void **)lide_c_util_malloc((size_t)(
                                        sizeof(void *) * 
                                        context->output_ports_count));

    if (context->buffer_start == NULL){
        free(context);
        return NULL;
    }
    for(i = 0; i < context->output_ports_count; i++){
        context->buffer_start[i] = (void *)lide_c_util_malloc((size_t)(
                                            context->capacity / 
                                            context->output_ports_count * 
                                            context->token_size));
        if (context->buffer_start[i] == NULL){
            free(context);
            return NULL;
        }
    }
    
    
    context->buffer_end = (void **)lide_c_util_malloc((size_t)(
                                        sizeof(void *) * 
                                        context->output_ports_count));

    if (context->buffer_end == NULL){
        free(context);
        return NULL;
    }
    for(i = 0; i < context->output_ports_count; i++){
        context->buffer_end[i] = ((char*)context->buffer_start[i]) + 
                ((context->capacity / context->output_ports_count - 1) * 
                context->token_size);
    }

    /*Read_pointer initialization*/
    context->read_pointer = NULL;
    context->read_pointer = (void **)lide_c_util_malloc(sizeof(void *) * 
                            context->output_ports_count);
                            
    /*Write_pointer initialization*/
    context->write_pointer = NULL;
    context->write_pointer = (void **)lide_c_util_malloc(sizeof(void *) * 
                            context->output_ports_count);
    
    /*Port initialization*/
    context->ports = NULL;
    context->ports = (lide_c_router_v2_port_pointer *)lide_c_util_malloc(
                                        sizeof(lide_c_router_v2_port_pointer)*
                                        context->ports_count);
    /* input port initialization*/
    for (i = 0; i < context->input_ports_count; i++){
        context->ports[i] = lide_c_router_v2_port_new(context, port_capacity, 
                                                INPUT_PORT, i, i);
    }
    /* output port initialization*/
    for (i = context->input_ports_count; i < context->ports_count; i++){
        tmp = i - context->input_ports_count;
        context->ports[i] = lide_c_router_v2_port_new(context, port_capacity, 
                                                OUTPUT_PORT, tmp, i);
    }

    /* methods */
    context->print = NULL;
    context->read = lide_c_router_v2_read;
    context->peek = lide_c_router_v2_peek;
    context->read_advance = lide_c_router_v2_read_advance;
    context->read_block = lide_c_router_v2_read_block;
    context->write = lide_c_router_v2_write;
    context->write_block = lide_c_router_v2_write_block;
    context->get_pop = lide_c_router_v2_population;
    //context->get_port_pop = lide_c_router_v2_port_population;
    context->get_cap = lide_c_router_v2_capacity;
    context->get_token_size = lide_c_router_v2_token_size;
    context->reset = lide_c_router_v2_reset;
    context->free = lide_c_router_v2_free;
    //context->get_port_index = lide_c_router_v2_port_get_index;
    context->get_port_withidx = lide_c_router_v2_port_get_router_port_pointer;
    context->get_port = lide_c_router_v2_get_port;
    //context->get_context = lide_c_router_v2_port_get_router_context;
    
    context->reset(context);
    return context;
}


void lide_c_router_v2_reset(lide_c_router_v2_context_pointer context){
    int i = 0;
    context->population = 0;
    //context->write_pointer = context->buffer_start;
    //for (i = 0; i<context->ports_count; i++){
    // may need some improvments
    for (i = 0; i<context->ports_count - context->input_ports_count; i++){
        context->read_pointer[i] = context->buffer_start[i];
        context->write_pointer[i] = context->buffer_start[i];
    }
    for(i = 0; i<context->ports_count; i++){
        context->ports[i]->population = 0;
    }
    return;
}

void lide_c_router_v2_set_token_printing(lide_c_router_v2_context_pointer router,
        lide_c_router_v2_f_token_printing f) {
    router->print = f;
    return;
}

void lide_c_router_v2_free(lide_c_router_v2_context_pointer router) {
    int i;
    if (router == NULL) {
        return;
    }
    if (router->buffer_start != NULL) {
        for (i = 0; i < router->output_ports_count; i++){
            if(router->buffer_start[i] != NULL)
                free(router->buffer_start[i]);
        }
        free(router->buffer_start);
    }
    if (router->read_pointer != NULL) {
        free(router->read_pointer);
    }
    
    if (router->write_pointer != NULL) {
        free(router->write_pointer);
    }

    if (router->ports != NULL) {
        for (i = 0; i < router->ports_count; i++){
            if(router->ports[i] != NULL)
                free(router->ports[i]);
        }
        free(router->ports);
    }   
    free(router);
    return;
}


int lide_c_router_v2_population(lide_c_router_v2_context_pointer router) {
    return router->population;
}

int lide_c_router_v2_capacity(lide_c_router_v2_context_pointer router) {
    return router->capacity;
}

int lide_c_router_v2_token_size(lide_c_router_v2_context_pointer router) {
    return router->token_size;
}


void lide_c_router_v2_read(lide_c_router_v2_port_pointer port, void *data){
    
    int index;
    int i,pop;
    lide_c_router_v2_context_pointer router;
    
    
    router = port->router_context;
    
    /*check whether it is an output port*/
    if (port->group_label == INPUT_PORT) {
        fprintf(stderr, "router read err: It is not output port in the \
                        router.\n");
        return;
    }  
    
    /* Make sure the router's related port is not empty. */
    if (router->population == 0 || port->population == 0) {
        fprintf(stderr, "router read err: population in the router is 0.\n");
        return;
    }
    
    /* index in the ports*/
    index = port->group_idx;
    /* The following assumes that sizeof(char) = 1. */
    memcpy(data, router->read_pointer[index], router->token_size);
    /* Advance the read pointer, using a circular buffering convention. */
    if (router->read_pointer[index] == router->buffer_end[index]) {
        router->read_pointer[index] = router->buffer_start[index];
    } else {
        router->read_pointer[index] = 
                ((char*)router->read_pointer[index]) + router->token_size;
    }
    
    /* Update the token count. */
    port->population--; 
    router->population--;
    /*
    pop = 0;
    for(i = router->input_ports_count; i < router->ports_count; i++){
        if (router->ports[i]->population > pop){
            pop = router->ports[i]->population;
        }
    }
    if(router->population != pop){
        router->population = pop;
        for(i = 0; i < router->input_ports_count; i++){
            router->ports[i]->population = pop;
        }
    }
    */
    return;
}

//TODO
void lide_c_router_v2_read_block(lide_c_router_v2_port_pointer port, void *data, 
                            int size){
    int index;
    int i,pop;
    /* Room (number of token spaces) "in front of the read pointer" within 
       the buffer. 
    */
    int room_in_front = 0;

    /* Breakdown in bytes of the amounts of data to read before and after,
       respectively, the read pointer wraps around the end of the buffer.
    */
    int part1_size = 0;
    int part2_size = 0;

    /* A cache of the FIFO token size. */
    int token_size = 0;
    
    lide_c_router_v2_context_pointer router;
    router = port->router_context;
    
    /*check whether it is an output port*/
    if (port->group_label == INPUT_PORT) {
        fprintf(stderr, "router read err: It is not output port.\n");
        return;
    } 
    
    /* Make sure the router's related port is not empty. */
    if (router->population < size || port->population < size) {
        fprintf(stderr, "router read err: population in the router is far \
                        from enough.\n");
        return;
    }

     
    
    /* index in the ports*/
    index = port->group_idx;

    token_size = router->token_size;
    room_in_front = ((((char*) (router->buffer_end[index])) - 
            ((char*)(router->read_pointer[index]))) / token_size) + 1;
    if (room_in_front >= size) {
        part1_size = size * token_size;
        part2_size = 0;
    } else {
        part1_size = room_in_front * token_size;
        part2_size = (size - room_in_front) * token_size; 
    }
    /* The following assumes that sizeof(char) = 1. */
    memcpy(data, router->read_pointer[index], part1_size);
    if (part2_size > 0) {
        data = ((char*)data) + part1_size;
        memcpy(data, router->buffer_start[index], part2_size);
        router->read_pointer[index] = ((char*)router->buffer_start[index]) + 
                                        part2_size;
    } else {
        if (size == room_in_front) {
            router->read_pointer[index] = router->buffer_start[index];
        } else {
            router->read_pointer[index] = 
                    ((char*)router->read_pointer[index]) + part1_size;
        }
    }
    /* Update Population for port and router */
    port->population -= size; 
    router->population -= size;
    /*
    pop = 0;
    for(i = router->input_ports_count; i < router->ports_count; i++){
        if (router->ports[i]->population > pop){
            pop = router->ports[i]->population;
        }
    }
    if(router->population != pop){
        router->population = pop;
        for(i = 0; i < router->input_ports_count; i++){
            router->ports[i]->population = pop;
        }
    }
    */
    return;
}
//TODO
void lide_c_router_v2_peek(lide_c_router_v2_port_pointer port, void *data){
    
}

//TODO
void lide_c_router_v2_read_advance(lide_c_router_v2_port_pointer port){
    
}


void lide_c_router_v2_write(lide_c_router_v2_port_pointer port, void *data){
    int i;
    lide_c_router_v2_context_pointer router;
    int en_buff_id;
    lide_c_router_v2_port_pointer wport;
    
    router = port->router_context;
    en_buff_id = router->en_buff_id;
    wport = router->ports[en_buff_id+router->input_ports_count];
    
    /*check whether it is an input port*/
    if (port->group_label == OUTPUT_PORT) {
        fprintf(stderr, "router read err: It is not input port.\n");
        return;
    } 

    /* Make sure the router is not full. */
    if (router->population >= router->capacity || 
        wport->population >= wport->capacity) {
        return;
    } 
    
     /* The following assumes that sizeof(char) = 1. */
    memcpy(router->write_pointer[en_buff_id], data, router->token_size);
    /* Advance the read pointer, using a circular buffering convention. */
    if (router->write_pointer[en_buff_id] == router->buffer_end[en_buff_id]) {
        router->write_pointer[en_buff_id] = router->buffer_start[en_buff_id];
    } else {
        router->write_pointer[en_buff_id] = 
                ((char*)router->write_pointer[en_buff_id]) + router->token_size;
    }
    /* Update the token count. */
    wport->population++;
    router->population++;
    /*    
    for(i = 0; i<router->ports_count; i++){
        router->ports[i]->population++;
    }
    */
    /* Update en_buff_id*/
    router->en_buff_id = (en_buff_id + 1)%router->output_ports_count;
    return;
}
//TODO update en_buff_id
void lide_c_router_v2_write_block(lide_c_router_v2_port_pointer port, void *data, 
                            int size){

    int i;
    void *tmp_data;
    /* Room (number of token spaces) "in front of the write pointer" within 
       the buffer. 
    */
    int room_in_front = 0;

    /* Breakdown in bytes of the amounts of data to write before and after,
       respectively, the write pointer wraps around the end of the buffer.
    */
    int part1_size = 0;
    int part2_size = 0;
    /* A cache of the FIFO token size. */
    int token_size = 0;
    
    lide_c_router_v2_context_pointer router;
    int en_buff_id;
    lide_c_router_v2_port_pointer wport;
    
    router = port->router_context;
    en_buff_id = router->en_buff_id;
    wport = router->ports[en_buff_id+router->input_ports_count];
    token_size = router->token_size;
    
    /*check whether it is an input port*/
    if (port->group_label == OUTPUT_PORT) {
        fprintf(stderr, "router read err: It is not input port.\n");
        return;
    } 
    
    /* Make sure the router is not full. */
    if ((router->population + size > router->capacity) || 
        (wport->population +  size > wport->capacity)) {
        return;
    } 
    
      
    /* The following assumes that sizeof(char) = 1. */
    room_in_front = ((((char*) (router->buffer_end[en_buff_id])) - 
            ((char*)(router->write_pointer[en_buff_id]))) / token_size) + 1;

    if (room_in_front >= size) {
        part1_size = size * router->token_size;
        part2_size = 0;
    } else {
        part1_size = room_in_front * router->token_size;
        part2_size = (size - room_in_front) * router->token_size; 
    }
    /* The following assumes that sizeof(char) = 1. */
    memcpy(router->write_pointer[en_buff_id], data, part1_size);
    
    if (part2_size > 0) {
        tmp_data = ((char*)data) + part1_size;
        memcpy(router->buffer_start[en_buff_id], tmp_data, part2_size);
        router->write_pointer[en_buff_id] = ((char*)router->buffer_start[en_buff_id]) + part2_size;
    } else {
        if (size == room_in_front) {
            router->write_pointer[en_buff_id] = router->buffer_start[en_buff_id];
        } else {
            router->write_pointer[en_buff_id] = ((char*)router->write_pointer[en_buff_id]) + part1_size;
        }
    }

    /* Update the token count. */
    wport->population += size;
    router->population += size;  
    /*
    for(i = 0; i<router->ports_count; i++){
        router->ports[i]->population += size;
    } 
    */
    /* Update en_buff_id*/
    router->en_buff_id = (en_buff_id + 1)%router->output_ports_count;
    
    return;
    
}


lide_c_router_v2_port_pointer lide_c_router_v2_get_port(
                        lide_c_router_v2_context_pointer router, 
                        int group_label, int group_idx){
    int port_index;
    port_index = router->input_ports_count * group_label + group_idx;
    return router->ports[port_index];
}


int lide_c_router_v2_port_population(lide_c_router_v2_port_pointer port){
    lide_c_router_v2_context_pointer router;
    int en_buff_id;
    router = port->router_context;
    en_buff_id = router->en_buff_id;
    if(port->group_label == INPUT_PORT){
        return router->ports[en_buff_id]->population;
    }else{
        return port->population;
    }
    
}

lide_c_router_v2_context_pointer lide_c_router_v2_port_get_router_context(
                                            lide_c_router_v2_port_pointer port){
    return port->router_context;
}

lide_c_router_v2_port_pointer lide_c_router_v2_port_get_router_port_pointer(
                                        lide_c_router_v2_context_pointer router, 
                                        int index){
    if(index < router -> ports_count && index >= 0){
        return router->ports[index];
    }else{
        fprintf(stderr, "port index is out of range.\n");
        return NULL;
    }
        
}

int lide_c_router_v2_port_get_index(lide_c_router_v2_port_pointer port){
    return port->index;
}

int lide_c_router_v2_context_population(
                            lide_c_router_v2_context_pointer router){
    return router->population;
}

int lide_c_router_v2_context_capacity(
                            lide_c_router_v2_context_pointer router){
    return router->capacity;
}


int lide_c_router_v2_port_capacity(lide_c_router_v2_port_pointer port){
    lide_c_router_v2_context_pointer router;
    int en_buff_id;
    router = port->router_context;
    en_buff_id = router->en_buff_id;
    if(port->group_label == INPUT_PORT){
        return router->ports[en_buff_id]->capacity;
    }else{
        return port->capacity;
    }
    
}


int lide_c_router_v2_get_enable_buffer_id(
                                    lide_c_router_v2_context_pointer router){
    return router->en_buff_id;
}
        









